if (randomizeInitValues)
{
    global P1 = Random (0.0001,1);
    global P2 = Random (0.0001,1);
    global W_1 = Random (0.0001,0.9999);
    global W_2 = Random (1.0001,5);
}
else
{
    global P1 = 1/3;
    global P2 = .5;
    global W_1 = .1;
    global W_2 = 5;
}

P1:<1;
P2:<1;
W_1 :< 1; W_1:>0.000000001;
W_2 :> 1; W_2:<100000;
categFreqMatrix = {{P1,(1-P1)*P2, (1-P1)*(1-P2)}} ;
categRateMatrix = {{W_1,1,W_2}};
category d = (3, categFreqMatrix , MEAN, ,categRateMatrix, 0, 1e25);			
