if (randomizeInitValues)
{
    global P1 = Random (0.0001,1);
    global W_1 = Random (0.0001,0.9999);
}
else
{
    global P1 = 1/2;
    global W_1 = .1;
}

P1:<1;
W_1 :< 1; W_1:>0.000000001;
categFreqMatrix = {{P1,(1-P1)}} ;
categRateMatrix = {{W_1,1}};
category d = (2, categFreqMatrix , MEAN, ,categRateMatrix, 0, 1e25);			
